@@ -1,23 +1,20 @@
 """Completers for Python code"""
-import re
-import sys
-import inspect
 import builtins
-import importlib
-import warnings
 import collections.abc as cabc
-from xonsh.parsers.completion_context import CompletionContext, PythonContext
+import inspect
+import re
+import warnings
 
-import xonsh.tools as xt
 import xonsh.lazyasd as xl
+import xonsh.tools as xt
 from xonsh.built_ins import XSH
-
 from xonsh.completers.tools import (
     CompleterResult,
+    RichCompletion,
     contextual_completer,
     get_filter_function,
-    RichCompletion,
 )
+from xonsh.parsers.completion_context import CompletionContext, PythonContext
 
 
 @xl.lazyobject
@@ -137,8 +134,8 @@ def complete_python(context: CompletionContext) -> CompleterResult:
     """
     Completes based on the contents of the current Python environment,
     the Python built-ins, and xonsh operators.
-    If there are no matches, split on common delimiters and try again.
     """
+    # If there are no matches, split on common delimiters and try again.
     if context.python is None:
         return None
 
@@ -241,7 +238,7 @@ def attr_complete(prefix, ctx, filter_func):
     prelen = len(prefix)
     for opt in opts:
         # check whether these options actually work (e.g., disallow 7.imag)
-        _expr = "{0}.{1}".format(expr, opt)
+        _expr = f"{expr}.{opt}"
         _val_, _ctx_ = _safe_eval(_expr, _ctx)
         if _val_ is None and _ctx_ is None:
             continue
@@ -276,51 +273,7 @@ def python_signature_complete(prefix, line, end, ctx, filter_func):
         return set()
     try:
         sig = inspect.signature(val)
-    except ValueError:
+    except (ValueError, TypeError):
         return set()
     args = {p + "=" for p in sig.parameters if filter_func(p, prefix)}
     return args
-
-
-@contextual_completer
-def complete_import(context: CompletionContext):
-    """
-    Completes module names and contents for "import ..." and "from ... import
-    ..."
-    """
-    if not (context.command and context.python):
-        # Imports are only possible in independent lines (not in `$()` or `@()`).
-        # This means it's python code, but also can be a command as far as the parser is concerned.
-        return None
-
-    command = context.command
-
-    if command.opening_quote:
-        # can't have a quoted import
-        return None
-
-    arg_index = command.arg_index
-    prefix = command.prefix
-    args = command.args
-
-    if arg_index == 1 and args[0].value == "from":
-        # completing module to import
-        return {"{} ".format(i) for i in complete_module(prefix)}
-    if arg_index >= 1 and args[0].value == "import":
-        # completing module to import
-        return complete_module(prefix)
-    if arg_index == 2 and args[0].value == "from":
-        return {RichCompletion("import", append_space=True)}
-    if arg_index > 2 and args[0].value == "from" and args[2].value == "import":
-        # complete thing inside a module
-        try:
-            mod = importlib.import_module(args[1].value)
-        except ImportError:
-            return set()
-        out = {i[0] for i in inspect.getmembers(mod) if i[0].startswith(prefix)}
-        return out
-    return set()
-
-
-def complete_module(prefix):
-    return {s for s in sys.modules if get_filter_function()(s, prefix)}
