@@ -1,32 +1,32 @@
 """Subprocess specification and related utilities."""
-import os
+import contextlib
+import inspect
 import io
+import os
+import pathlib
 import re
-import sys
 import shlex
 import signal
-import inspect
-import pathlib
-import builtins
+import stat
 import subprocess
-import contextlib
+import sys
 
-import xonsh.tools as xt
-import xonsh.lazyasd as xl
-import xonsh.platform as xp
 import xonsh.environ as xenv
-import xonsh.lazyimps as xli
 import xonsh.jobs as xj
-
-from xonsh.procs.readers import ConsoleParallelReader
-from xonsh.procs.posix import PopenThread
-from xonsh.procs.proxies import ProcProxy, ProcProxyThread
+import xonsh.lazyasd as xl
+import xonsh.lazyimps as xli
+import xonsh.platform as xp
+import xonsh.tools as xt
+from xonsh.built_ins import XSH
 from xonsh.procs.pipelines import (
-    pause_call_resume,
+    STDOUT_CAPTURE_KINDS,
     CommandPipeline,
     HiddenCommandPipeline,
-    STDOUT_CAPTURE_KINDS,
+    pause_call_resume,
 )
+from xonsh.procs.posix import PopenThread
+from xonsh.procs.proxies import ProcProxy, ProcProxyThread
+from xonsh.procs.readers import ConsoleParallelReader
 
 
 @xl.lazyobject
@@ -39,13 +39,19 @@ def is_app_execution_alias(fname):
     Here we try to detect if a file is an app execution alias.
     """
     fname = pathlib.Path(fname)
-    return not os.path.exists(fname) and fname.name in os.listdir(fname.parent)
+    try:
+        return fname.stat().st_reparse_tag == stat.IO_REPARSE_TAG_APPEXECLINK
+
+    # os.stat().st_reparse_tag is python 3.8+, and os.stat(app_exec_alias) throws OSError for <= 3.7
+    # so use old method as fallback
+    except (AttributeError, OSError):
+        return not os.path.exists(fname) and fname.name in os.listdir(fname.parent)
 
 
 def _is_binary(fname, limit=80):
     try:
         with open(fname, "rb") as f:
-            for i in range(limit):
+            for _ in range(limit):
                 char = f.read(1)
                 if char == b"\0":
                     return True
@@ -101,7 +107,7 @@ def get_script_subproc_command(fname, args):
         # Windows can execute various filetypes directly
         # as given in PATHEXT
         _, ext = os.path.splitext(fname)
-        if ext.upper() in builtins.__xonsh__.env.get("PATHEXT"):
+        if ext.upper() in XSH.env.get("PATHEXT"):
             return [fname] + args
     # find interpreter
     with open(fname, "rb") as f:
@@ -157,16 +163,12 @@ def _REDIR_OUT():
 
 @xl.lazyobject
 def _E2O_MAP():
-    return frozenset(
-        {"{}>{}".format(e, o) for e in _REDIR_ERR for o in _REDIR_OUT if o != ""}
-    )
+    return frozenset({f"{e}>{o}" for e in _REDIR_ERR for o in _REDIR_OUT if o != ""})
 
 
 @xl.lazyobject
 def _O2E_MAP():
-    return frozenset(
-        {"{}>{}".format(o, e) for e in _REDIR_ERR for o in _REDIR_OUT if o != ""}
-    )
+    return frozenset({f"{o}>{e}" for e in _REDIR_ERR for o in _REDIR_OUT if o != ""})
 
 
 def _is_redirect(x):
@@ -177,7 +179,7 @@ def safe_open(fname, mode, buffering=-1):
     """Safely attempts to open a file in for xonsh subprocs."""
     # file descriptors
     try:
-        return io.open(fname, mode, buffering=buffering)
+        return open(fname, mode, buffering=buffering)
     except PermissionError:
         raise xt.XonshError(f"xonsh: {fname}: permission denied")
     except FileNotFoundError:
@@ -355,6 +357,8 @@ class SubprocSpec:
         # pure attrs
         self.args = list(cmd)
         self.alias = None
+        self.alias_name = None
+        self.alias_stack = XSH.env.get("__ALIAS_STACK", "").split(":")
         self.binary_loc = None
         self.is_proxy = False
         self.background = False
@@ -440,10 +444,12 @@ class SubprocSpec:
         event_name = self._cmd_event_name()
         self._pre_run_event_fire(event_name)
         kwargs = {n: getattr(self, n) for n in self.kwnames}
-        self.prep_env(kwargs)
         if callable(self.alias):
+            kwargs["env"] = self.env or {}
+            kwargs["env"]["__ALIAS_NAME"] = self.alias_name or ""
             p = self.cls(self.alias, self.cmd, **kwargs)
         else:
+            self.prep_env_subproc(kwargs)
             self.prep_preexec_fn(kwargs, pipeline_group=pipeline_group)
             self._fix_null_cmd_bytes()
             p = self._run_binary(kwargs)
@@ -455,9 +461,16 @@ class SubprocSpec:
         return p
 
     def _run_binary(self, kwargs):
+        if not self.cmd[0]:
+            raise xt.XonshError("xonsh: subprocess mode: command is empty")
+        bufsize = 1
         try:
-            bufsize = 1
-            p = self.cls(self.cmd, bufsize=bufsize, **kwargs)
+            if xp.ON_WINDOWS and self.binary_loc is not None:
+                # launch process using full paths (https://bugs.python.org/issue8557)
+                cmd = [self.binary_loc] + self.cmd[1:]
+            else:
+                cmd = self.cmd
+            p = self.cls(cmd, bufsize=bufsize, **kwargs)
         except PermissionError:
             e = "xonsh: subprocess mode: permission denied: {0}"
             raise xt.XonshError(e.format(self.cmd[0]))
@@ -468,17 +481,17 @@ class SubprocSpec:
                     return self.cls(
                         ["man", cmd0.rstrip("?")], bufsize=bufsize, **kwargs
                     )
-            e = "xonsh: subprocess mode: command not found: {0}".format(cmd0)
-            env = builtins.__xonsh__.env
-            sug = xt.suggest_commands(cmd0, env, builtins.aliases)
+            e = f"xonsh: subprocess mode: command not found: {cmd0}"
+            env = XSH.env
+            sug = xt.suggest_commands(cmd0, env)
             if len(sug.strip()) > 0:
-                e += "\n" + xt.suggest_commands(cmd0, env, builtins.aliases)
+                e += "\n" + xt.suggest_commands(cmd0, env)
             raise xt.XonshError(e)
         return p
 
-    def prep_env(self, kwargs):
+    def prep_env_subproc(self, kwargs):
         """Prepares the environment to use in the subprocess."""
-        with builtins.__xonsh__.env.swap(self.env) as env:
+        with XSH.env.swap(self.env) as env:
             denv = env.detype()
         if xp.ON_WINDOWS:
             # Over write prompt variable as xonsh's $PROMPT does
@@ -490,9 +503,9 @@ class SubprocSpec:
         """Prepares the 'preexec_fn' keyword argument"""
         if not xp.ON_POSIX:
             return
-        if not builtins.__xonsh__.env.get("XONSH_INTERACTIVE"):
+        if not XSH.env.get("XONSH_INTERACTIVE"):
             return
-        if pipeline_group is None or xp.ON_WSL:
+        if pipeline_group is None or xp.ON_WSL1:
             # If there is no pipeline group
             # or the platform is windows subsystem for linux (WSL)
             xonsh_preexec_fn = no_pg_xonsh_preexec_fn
@@ -514,6 +527,8 @@ class SubprocSpec:
         # escape them just in case.
         cmd = self.cmd
         for i in range(len(cmd)):
+            if callable(cmd[i]):
+                raise Exception(f"The command contains callable argument: {cmd[i]}")
             cmd[i] = cmd[i].replace("\0", "\\0")
 
     def _cmd_event_name(self):
@@ -525,14 +540,14 @@ class SubprocSpec:
             return os.path.basename(self.binary_loc)
 
     def _pre_run_event_fire(self, name):
-        events = builtins.events
+        events = XSH.builtins.events
         event_name = "on_pre_spec_run_" + name
         if events.exists(event_name):
             event = getattr(events, event_name)
             event.fire(spec=self)
 
     def _post_run_event_fire(self, name, proc):
-        events = builtins.events
+        events = XSH.builtins.events
         event_name = "on_post_spec_run_" + name
         if events.exists(event_name):
             event = getattr(events, event_name)
@@ -548,6 +563,8 @@ class SubprocSpec:
         modifications and adjustments based on the actual cmd that
         was received.
         """
+        if not cmd:
+            raise xt.XonshError("xonsh: subprocess mode: command is empty")
         # modifications that do not alter cmds may come before creating instance
         spec = kls(cmd, cls=cls, **kwargs)
         # modifications that alter cmds must come after creating instance
@@ -564,7 +581,7 @@ class SubprocSpec:
         return spec
 
     def redirect_leading(self):
-        """Manage leading redirects such as with '< input.txt COMMAND'. """
+        """Manage leading redirects such as with '< input.txt COMMAND'."""
         while len(self.cmd) >= 3 and self.cmd[0] == "<":
             self.stdin = safe_open(self.cmd[1], "r")
             self.cmd = self.cmd[2:]
@@ -587,17 +604,29 @@ class SubprocSpec:
     def resolve_alias(self):
         """Sets alias in command, if applicable."""
         cmd0 = self.cmd[0]
+
+        if cmd0 in self.alias_stack:
+            # Disabling the alias resolving to prevent infinite loop in call stack
+            # and futher using binary_loc to resolve the alias name.
+            self.alias = None
+            return
+
         if callable(cmd0):
             alias = cmd0
         else:
-            alias = builtins.aliases.get(cmd0, None)
+            alias = XSH.aliases.get(cmd0, None)
+            if alias is not None:
+                self.alias_name = cmd0
         self.alias = alias
 
     def resolve_binary_loc(self):
         """Sets the binary location"""
         alias = self.alias
         if alias is None:
-            binary_loc = xenv.locate_binary(self.cmd[0])
+            cmd0 = self.cmd[0]
+            binary_loc = xenv.locate_binary(cmd0)
+            if binary_loc == cmd0 and cmd0 in self.alias_stack:
+                raise Exception(f'Recursive calls to "{cmd0}" alias.')
         elif callable(alias):
             binary_loc = None
         else:
@@ -610,12 +639,12 @@ class SubprocSpec:
             self.alias is None
             and self.binary_loc is None
             and len(self.cmd) == 1
-            and builtins.__xonsh__.env.get("AUTO_CD")
+            and XSH.env.get("AUTO_CD")
             and os.path.isdir(self.cmd[0])
         ):
             return
         self.cmd.insert(0, "cd")
-        self.alias = builtins.aliases.get("cd", None)
+        self.alias = XSH.aliases.get("cd", None)
 
     def resolve_executable_commands(self):
         """Resolve command executables, if applicable."""
@@ -646,7 +675,7 @@ class SubprocSpec:
         if not callable(alias):
             return
         self.is_proxy = True
-        env = builtins.__xonsh__.env
+        env = XSH.env
         thable = env.get("THREAD_SUBPROCS") and getattr(
             alias, "__xonsh_threadable__", True
         )
@@ -676,7 +705,7 @@ class SubprocSpec:
 
 
 def _safe_pipe_properties(fd, use_tty=False):
-    """Makes sure that a pipe file descriptor properties are sane."""
+    """Makes sure that a pipe file descriptor properties are reasonable."""
     if not use_tty:
         return
     # due to some weird, long standing issue in Python, PTYs come out
@@ -701,7 +730,7 @@ def _safe_pipe_properties(fd, use_tty=False):
 
 
 def _update_last_spec(last):
-    env = builtins.__xonsh__.env
+    env = XSH.env
     captured = last.captured
     last.last_in_pipeline = True
     if not captured:
@@ -710,9 +739,10 @@ def _update_last_spec(last):
     if callable_alias:
         pass
     else:
-        cmds_cache = builtins.__xonsh__.commands_cache
+        cmds_cache = XSH.commands_cache
         thable = (
             env.get("THREAD_SUBPROCS")
+            and (captured != "hiddenobject" or env.get("XONSH_CAPTURE_ALWAYS"))
             and cmds_cache.predict_threadable(last.args)
             and cmds_cache.predict_threadable(last.cmd)
         )
@@ -736,9 +766,9 @@ def _update_last_spec(last):
         r, w = os.pipe()
         last.stdout = safe_open(w, "wb")
         last.captured_stdout = safe_open(r, "rb")
-    elif builtins.__xonsh__.stdout_uncaptured is not None:
+    elif XSH.stdout_uncaptured is not None:
         last.universal_newlines = True
-        last.stdout = builtins.__xonsh__.stdout_uncaptured
+        last.stdout = XSH.stdout_uncaptured
         last.captured_stdout = last.stdout
     elif xp.ON_WINDOWS and not callable_alias:
         last.universal_newlines = True
@@ -754,12 +784,14 @@ def _update_last_spec(last):
     # set standard error
     if last.stderr is not None:
         pass
+    elif captured == "stdout":
+        pass
     elif captured == "object":
         r, w = os.pipe()
         last.stderr = safe_open(w, "w")
         last.captured_stderr = safe_open(r, "r")
-    elif builtins.__xonsh__.stderr_uncaptured is not None:
-        last.stderr = builtins.__xonsh__.stderr_uncaptured
+    elif XSH.stderr_uncaptured is not None:
+        last.stderr = XSH.stderr_uncaptured
         last.captured_stderr = last.stderr
     elif xp.ON_WINDOWS and not callable_alias:
         last.universal_newlines = True
@@ -809,18 +841,29 @@ def cmds_to_specs(cmds, captured=False, envs=None):
             specs[i].background = True
         else:
             raise xt.XonshError(f"unrecognized redirect {redirect!r}")
+
     # Apply boundary conditions
+    if not XSH.env.get("XONSH_CAPTURE_ALWAYS"):
+        # Make sure sub-specs are always captured.
+        # I.e. ![some_alias | grep x] $(some_alias)
+        specs_to_capture = specs if captured in STDOUT_CAPTURE_KINDS else specs[:-1]
+        for spec in specs_to_capture:
+            if spec.env is None:
+                spec.env = {"XONSH_CAPTURE_ALWAYS": True}
+            else:
+                spec.env.setdefault("XONSH_CAPTURE_ALWAYS", True)
+
     _update_last_spec(specs[-1])
     return specs
 
 
 def _should_set_title(captured=False):
-    env = builtins.__xonsh__.env
+    env = XSH.env
     return (
         env.get("XONSH_INTERACTIVE")
         and not env.get("XONSH_STORE_STDOUT")
         and captured not in STDOUT_CAPTURE_KINDS
-        and builtins.__xonsh__.shell is not None
+        and XSH.shell is not None
     )
 
 
@@ -837,8 +880,12 @@ def run_subproc(cmds, captured=False, envs=None):
 
     Lastly, the captured argument affects only the last real command.
     """
-    if builtins.__xonsh__.env.get("XONSH_TRACE_SUBPROC"):
-        print(f"TRACE SUBPROC: {cmds}", file=sys.stderr)
+    if XSH.env.get("XONSH_TRACE_SUBPROC", False):
+        tracer = XSH.env.get("XONSH_TRACE_SUBPROC_FUNC")
+        if callable(tracer):
+            tracer(cmds, captured=captured)
+        else:
+            print(f"TRACE SUBPROC: {cmds}, captured={captured}", file=sys.stderr)
 
     specs = cmds_to_specs(cmds, captured=captured, envs=envs)
     captured = specs[-1].captured
@@ -861,25 +908,25 @@ def run_subproc(cmds, captured=False, envs=None):
         )
     if _should_set_title(captured=captured):
         # set title here to get currently executing command
-        pause_call_resume(proc, builtins.__xonsh__.shell.settitle)
+        pause_call_resume(proc, XSH.shell.settitle)
     else:
         # for some reason, some programs are in a stopped state when the flow
         # reaches this point, hence a SIGCONT should be sent to `proc` to make
         # sure that the shell doesn't hang. This `pause_call_resume` invocation
         # does this
         pause_call_resume(proc, int)
-    # create command or return if backgrounding.
-    if background:
+    # now figure out what we should return
+    if captured == "object":
+        return command  # object can be returned even if backgrounding
+    elif captured == "hiddenobject":
+        if not background:
+            command.end()
+        return command
+    elif background:
         return
-    # now figure out what we should return.
-    if captured == "stdout":
+    elif captured == "stdout":
         command.end()
         return command.output
-    elif captured == "object":
-        return command
-    elif captured == "hiddenobject":
-        command.end()
-        return command
     else:
         command.end()
         return
