@@ -1,7 +1,6 @@
-# -*- coding: utf-8 -*-
 """Prompt formatter for simple version control branches"""
 # pylint:disable=no-member, invalid-name
-
+import contextlib
 import os
 import pathlib
 import queue
@@ -30,13 +29,18 @@ def _run_git_cmd(cmd):
 
 
 def _get_git_branch(q):
-    try:
-        cmd = ["git", "rev-parse", "--abbrev-ref", "HEAD"]
-        branch = xt.decode_bytes(_run_git_cmd(cmd))
-    except (subprocess.CalledProcessError, OSError):
-        q.put(None)
-    else:
-        q.put(branch.splitlines()[0] if branch else None)
+    # from https://git-blame.blogspot.com/2013/06/checking-current-branch-programatically.html
+    for cmds in [
+        "git symbolic-ref --short HEAD",
+        "git show-ref --head -s --abbrev",  # in detached mode return sha1
+    ]:
+        with contextlib.suppress(subprocess.CalledProcessError, OSError):
+            branch = xt.decode_bytes(_run_git_cmd(cmds.split()))
+            if branch:
+                q.put(branch.splitlines()[0])
+                return
+    # all failed
+    q.put(None)
 
 
 def get_git_branch():
@@ -96,7 +100,7 @@ def get_hg_branch(root=None):
         # get branch name
         branch_path = root / ".hg" / "branch"
         if branch_path.exists():
-            with open(branch_path, "r") as branch_file:
+            with open(branch_path) as branch_file:
                 branch = branch_file.read().strip()
         else:
             branch = "default"
